from .config import CONFIG_PATH, APP_DIR, set_table_name, DEBUG, get_config
import click
import os
import shutil
import importlib.resources
from datetime import datetime
import subprocess


def d_print(string: str) -> None:
    # Print if debug mode enabled

    if DEBUG:
        click.echo(f"DEBUG: {string}")


def hide_cursor():
    print("\033[?25l", end="")


def show_cursor():
    print("\033[?25h", end="")


def has_custom_configs() -> bool:
    if get_config():
        d_print(f"config len: {len(get_config())}")
        return len(get_config()) > 1
    else:
        raise Exception("Config not found!")


def list_versions():
    # Clean this up later whatever
    config_map = {}
    i = 1
    if len(get_config()) > 1:
        print(f"({len(get_config()) - 1}) custom configs detected:")
        for v in get_config():
            if v != "default":
                config_map[str(i)] = v
                print(f"  {i}:  {v}")
                i += 1
        return config_map

    return None


def reset_config():
    # Check if config exists yet
    os.makedirs(APP_DIR, exist_ok=True)

    if os.path.exists(APP_DIR) and os.path.exists(CONFIG_PATH):
        default_config = importlib.resources.files("odoo_soup").joinpath(
            "default_soups.toml"
        )

        with importlib.resources.as_file(default_config) as conf:
            shutil.copyfile(conf, CONFIG_PATH)

        click.echo(f"Replaced {CONFIG_PATH}")

    else:
        # If theres no config file there just make one instead
        init_dirs(False)


def init_dirs(show=True):
    # Make app dir
    os.makedirs(APP_DIR, exist_ok=True)

    # Check if config exists yet
    if os.path.exists(APP_DIR) and os.path.exists(CONFIG_PATH):
        if show:
            click.echo(f"-> Using config file: {CONFIG_PATH}")

    else:
        if show:
            click.echo(f"No config file found at {CONFIG_PATH}, creating now...")

        default_config = importlib.resources.files("odoo_soup").joinpath(
            "default_soups.toml"
        )

        with importlib.resources.as_file(default_config) as conf:
            shutil.copyfile(conf, CONFIG_PATH)

        click.echo(f"Created {CONFIG_PATH}")


def db_exists():
    r = subprocess.run(
        "psql -lqt | cut -d \\| -f 1 | grep -qw 'odoo-soup'",
        shell=True,
        check=False,
        capture_output=True,
        text=True,
    )
    return not r.returncode


def create_db():
    click.echo("-> Database 'odoo-soup' does not exist. Creating it now...")
    subprocess.run(
        "createdb odoo-soup",
        shell=True,
        check=True,
        text=True,
    )


def init(table):
    # Init/check config and app dir in place
    init_dirs()

    # Forward tablename arg to config
    set_table_name(table)

    # Check if the odoo-soup db is made
    if not db_exists():
        create_db()


def scan_log_files() -> list[str]:
    # Read through the provided directory
    # Find any .gz files
    files = []

    for entry in os.scandir(os.getcwd()):
        if entry.is_file():
            # Identify if it is of a type we care about
            ext = entry.name.split(".")
            ext = ext[len(ext) - 1]

            if ext == "gz":
                # Add name and path to files list
                files.append((entry.name, entry.path))

    # Break if there were no files found
    if not len(files):
        raise FileNotFoundError(
            "No .gz logfiles were found in CWD! Make sure you run this in the folder where your logs are"
        )
    return files


def print_list_tuple(lt: list[tuple]) -> int:
    # Prints a list of numbered options and returns length
    for i in range(len(lt)):
        click.echo(f"  {i + 1}: {lt[i][0]}")

    click.echo()
    return len(lt)



import click
from .utils import (
    init,
    scan_log_files,
    print_list_tuple,
    d_print,
    has_custom_configs,
    list_versions,
    init_dirs,
    reset_config,
)
from .compiler import parse_config
from .converter import convert_to_sql
from . import config

from magicprompt import prompt
import subprocess
import os
import gzip
import shutil


@click.group()
def soup():
    pass


@soup.command()
def nuke():
    """
    Delete the 'odoo-soup' DB (deletes ALL imported log tables)
    """
    if click.confirm(
        "Are you sure? This will delete ALL imported logs within 'odoo-soup' DB"
    ):
        click.echo("-> Nuking 'odoo-soup'...")
        try:
            subprocess.run(
                "dropdb odoo-soup",
                shell=True,
                check=True,
                text=True,
            )
            click.echo("   Done!")
        except:
            raise Exception(
                "Could not drop 'odoo-soup'. It either doesn't exist or there are connections to it which must be closed."
            )


@soup.command()
def edit():
    """
    Open soup config in nano
    """
    init_dirs(False)
    click.edit(filename=config.CONFIG_PATH)


@soup.command()
def show():
    """
    Show soup config path
    """
    init_dirs(False)
    click.echo(config.CONFIG_PATH)


@soup.command()
def reset():
    """
    Reset soup config file
    """
    reset_config()


@soup.command()
@click.option(
    "--keep",
    is_flag=True,
    help="Keep the original file alongside compressed output",
)
@click.argument("file")
def zip(keep, file):
    """
    Convert a plaintext log to .gz for import
    """
    file_path = os.path.join(os.getcwd(), file)

    # Check that the file exists
    if os.path.exists(file_path):
        # Check that it is not .gz
        if file_path[-3:] == ".gz":
            raise TypeError(f"File {file_path} is already .gz archive")

        # Make the archive
        click.echo(f"-> Compressing {file}...")
        with open(file_path, "rb") as f:
            with gzip.open(f"{file_path}.gz", "wb") as f_out:
                shutil.copyfileobj(f, f_out)

        # Remove original by default
        if not keep:
            os.remove(file_path)

        click.echo(f"   Done: {file}.gz")

        return file_path + ".gz"

    else:
        raise FileNotFoundError(f"File {file_path} does not exist!")


@soup.command()
@click.option(
    "--keep",
    is_flag=True,
    help="Keep the original file alongside uncompressed output",
)
@click.argument("file")
def unzip(keep, file):
    """
    Convert a .gz log to plaintext
    """
    file_path = os.path.join(os.getcwd(), file)

    # Check that the file exists
    if os.path.exists(file_path):
        # Check that it is not .gz
        if file_path[-3:] == ".gz":
            # Make the archive
            click.echo(f"-> Decompressing {file}...")
            with gzip.open(file_path, "rb") as f:
                with open(file_path[:-3], "wb") as f_out:
                    shutil.copyfileobj(f, f_out)

            # Remove original by default
            if not keep:
                os.remove(file_path)

            click.echo(f"   Done: {file[:-3]}")

            return file_path[:-3]

        else:
            raise TypeError(f"File {file_path} is not a .gz archive")

    else:
        raise FileNotFoundError(f"File {file_path} does not exist!")


@soup.command()
@click.option(
    "--nozip",
    is_flag=True,
    help="Skip converting filtered file to .gz",
)
@click.argument("file")
@click.argument("database")
def filter(nozip, file, database):
    """
    Filter a SAAS log to a specific Odoo database
    """
    file_path = os.path.join(os.getcwd(), file)

    # Check that the file exists
    if os.path.exists(file_path):
        # If it is compressed, extract the file
        if file_path[-3:] == ".gz":
            file_path = unzip.callback(True, file)
            file = file[:-3]

        click.echo("-> Filtering...")
        r = subprocess.run(
            f"grep '{database}' {file_path} >> {file_path}.filtered",
            shell=True,
            text=True,
        )

        if r.returncode > 1:
            raise Exception("Could not filter that file. Idk bro.")

        # Zip the new filtered file for import
        if not nozip:
            zip.callback(False, file + ".filtered")

        else:
            click.echo(f"   Done: {file}.filtered")


# @click.option(
#     "--dirty",
#     is_flag=True,
#     help="Include ALL unconfigured line types as text records",
# )
@soup.command()
@click.argument("table")
def make(table):
    """
    Convert a .gz logfile to SQL table
    """

    # Init/check the dirs and db
    init(table)

    # Get the log files from CWD
    log_files = scan_log_files()

    click.clear()

    # USER SELECTS LOGFILE
    click.echo("Log files in current directory:\n")
    numFiles = print_list_tuple(log_files)

    resp = (
        prompt(
            "Select a logfile to convert",
            numOnly=1,
            clearAfterResponse=1,
            validators=[lambda x: 0 < int(x) <= numFiles],
        )
        - 1
    )
    click.clear()

    d_print(f"resp: {resp}")
    d_print(f"config: {config.get_config()}")

    # USER SELECTS CUSTOM CONFIG
    if has_custom_configs():
        options = list_versions()

        vers = prompt(
            "Choose a configuration (ENTER for default)",
            capPrompt=False,
            castAnswer=False,
            notEmpty=False,
            clearAfterResponse=1,
            validators=[lambda x: x in options or x == ""],
        )

        version = options.get(vers, "default")
    else:
        version = "default"

    # APP GO BRRR
    click.echo(f"Converting {log_files[resp][0]} to SQL table...\n")

    # Compile mapping
    config.set_mapping(parse_config(version))
    d_print(f"compiled mapping: {config.MAPPING}")

    # Convert to sql
    skipped, stats = convert_to_sql(log_files[resp][1])

    click.clear()
    click.echo(
        f"Success! Created table: {config.TABLE_NAME} in DB: odoo-soup ({int(stats['time'] * 1000) / 1000.0} s)\n"
    )

    d_print(
        f"File length: {stats['raw_length']} | Traceback/junk lines: {stats['junk_skipped']} | Imported log lines: {stats['table_len']}"
    )

    click.echo(
        "The following line types were skipped due to not being present in your config (sorted by freq):"
    )

    for s in skipped:
        click.echo(f"  {s}")

    click.echo(
        f"\nTotal lines skipped: {stats['log_lines_skipped']} ({100.0 - (int((stats['log_lines_skipped'] / (stats['table_len'] - stats['junk_skipped'])) * 10000) / 100.0)}% coverage)"
    )


if __name__ == "__main__":
    soup()



from .config import get_config


def parse_fields(fieldsList: list) -> dict:
    fieldDict = {}
    for field in fieldsList:
        # Fields have optional string slicing indexes
        sliceStart = None
        sliceEnd = None

        # Format in config is arr with start and end pos, end pos optional
        if len(field) == 3:
            sliceStart = field[2][0]
            if len(field[2]) == 2:
                sliceEnd = field[2][1]

        fieldDict[field[0]] = {
            "index": field[1],
            "sliceStart": sliceStart,
            "sliceEnd": sliceEnd,
        }

    return fieldDict


def parse_config(version: str):
    # It is easier for user to configure logline types by alias rather than the search pattern, so some changes must be made here to read efficiently
    # Ensure that there are expected members of config for parsing log lines
    try:
        defaultIndexes = get_config()["default"]["default"]
    except KeyError:
        raise KeyError("Config is missing default parameters")

    """
    1. build a dict of search patterns mapped to alias, with the base default fields for all types of log lines
    2. update that dict with the default type specific field info
    2. if a non default version was configured, update the dict with the version specific info as well. return default config otherwise
    """

    # Populate the type mapping with the default options and overwrite any version specific things
    typeDefs = {}

    defaultTypes = get_config()["default"]
    defaultFields = parse_fields(defaultIndexes["fields"])
    # print(defaultTypes.items())

    for type, definition in defaultTypes.items():
        if type == "default":
            continue

        # Swap the alias and pattern for O(1) search while allowing easier config writing
        # Also add the default base mappings in
        temp = {"alias": type, "fields": defaultFields.copy()}

        # Check if there are type specific default mappings
        fields = definition.get("fields")

        if fields:
            fields = parse_fields(fields)
            temp["fields"].update(fields)

        typeDefs[definition["pattern"]] = temp

    # Now we overwrite the default options with version specific options if chosen by user

    if version != "default":
        versionSpecificTypes = get_config()[version]

        for type, definition in versionSpecificTypes.items():
            if type == "default":
                continue
            # Look up the pattern among existing mappings
            pattern = definition.get("pattern")

            # If the pattern is not a key in typedefs, that means we are redefining this mapping. So we need to look up the alias and swap to this search pattern
            if pattern not in typeDefs:
                # Find the key with matching alias
                for k, v in typeDefs.items():
                    if v["alias"] == type:
                        # If we found a record with matching alias, copy its contents to a new record with the new pattern, and delete old one
                        if pattern:
                            typeDefs[pattern] = v.copy()
                            del typeDefs[k]
                        else:
                            # Cover case where pattern was None so that a None key is not made
                            pattern = k
                        break

                # Add it in if it is truly not present at all
                if pattern is not None:
                    typeDefs[pattern] = {"alias": type, "fields": defaultFields.copy()}

            # Now that it is guaranteed to be in, we will update the record with the version specific info
            fields = definition.get("fields")
            if fields:
                definition["fields"] = parse_fields(fields)
                typeDefs[pattern]["fields"].update(definition["fields"])

    typeDefs["default"] = defaultFields

    return typeDefs



import click
import os
import tomllib
from .models import create_log_model
from sqlalchemy.orm import declarative_base
from sqlalchemy import create_engine
from datetime import datetime


BASE = declarative_base()

# Identify paths
APP_DIR = click.get_app_dir("odoo_soup")
CONFIG_PATH = os.path.join(APP_DIR, "soups.toml")


CONFIG = None
DEBUG = False
TABLE_MODEL = None
TABLE_NAME = None
MAPPING = None
ENGINE = create_engine("postgresql:///odoo-soup")


def load_config(path: str) -> dict:
    with open(path, "rb") as f:
        # click.echo(f"Parsing config at {path}...")
        try:
            config = tomllib.load(f)
            return config
        except:
            raise Exception(
                f"Error parsing {path}. Verify that there are no configuration typos"
            )


# For lazyloading stuff
def get_config() -> dict:
    global CONFIG

    if CONFIG is None:
        CONFIG = load_config(CONFIG_PATH)

    return CONFIG


def set_table_name(name):
    global TABLE_NAME
    TABLE_NAME = name


def set_mapping(mapping):
    global MAPPING
    MAPPING = mapping


def get_mapping():
    global MAPPING
    return MAPPING


def get_table_model():
    global TABLE_MODEL

    if TABLE_MODEL is None:
        TABLE_MODEL = create_log_model(TABLE_NAME, BASE)

    return TABLE_MODEL


TYPES = {
    "date": datetime,
    "level": str,
    "origin": str,
    "ip": str,
    "http": str,
    "route": str,
    "code": int,
    "time": float,
    "user": str,
    "object": str,
    "records": str,
    "text": str,
}



from sqlalchemy import Column, Integer, Text, Double, DateTime, String


# Factory for logline model so that name can be dynamic
def create_log_model(table_name: str, base):
    class Log(base):
        __tablename__ = table_name

        id = Column(Integer, primary_key=True)
        date = Column(DateTime, nullable=False)
        level = Column(String(15))
        origin = Column(Text)
        type = Column(String(64))
        ip = Column(String(128))
        http = Column(String(10))
        route = Column(Text)
        code = Column(Integer)
        time = Column(Double)
        user = Column(Text)
        object = Column(Text)
        records = Column(Text)
        text = Column(Text)

    return Log



from datetime import datetime
from .config import BASE, get_mapping, ENGINE, TYPES, get_table_model
from .utils import show_cursor, hide_cursor
from sqlalchemy.orm import sessionmaker
import gzip
import click
import os
import click_spinner
import time


def check_date(date: str):
    try:
        format = "%Y-%m-%d %H:%M:%S"
        date = datetime.strptime(date, format)
        return date
    except:
        return None


def condense(skipped) -> list[str]:
    freqMap = {}

    # Map frequencies of skipped lines
    for s in skipped:
        freqMap[s] = freqMap.get(s, 0) + 1

    # Sort the dict
    sort = sorted(freqMap.items(), key=lambda kv: (kv[1], kv[0]), reverse=True)

    pretty = []

    # Make the frequencies pretty
    for k, v in sort:
        pretty.append(f"{k} -> ({v})")

    return pretty


def apply_trim(str, field, lineType=None):
    # Again, type will have to be done blindly
    if field == "type":
        sS, sE = (
            get_mapping()["default"]["type"]["sliceStart"],
            get_mapping()["default"]["type"]["sliceEnd"],
        )
    else:
        sS, sE = (
            get_mapping()[lineType]["fields"][field]["sliceStart"],
            get_mapping()[lineType]["fields"][field]["sliceEnd"],
        )

    # There should always be slice start, be it 0 or something specific
    if sS is None and sE is None:
        return str
    if sE is None or sE == 0:
        return str[sS:]
    return str[sS:sE]


def from_line(split_line, field, lineType=None):
    # type needs to be generic determined for any line
    if field == "type":
        # print(get_mapping())
        raw = split_line[get_mapping()["default"]["type"]["index"]]
        return apply_trim(raw, field)

    f = get_mapping()[lineType]["fields"].get(field)
    if f:
        try:
            raw = split_line[f["index"]]
            raw = apply_trim(raw, field, lineType)
            # Cast the contents to the expected type for sql
            raw = TYPES[field](raw)
            return raw
        except:
            return None


def process_line(line: str) -> tuple:
    line = line.strip()

    skeletor = {
        "date": None,
        "level": None,
        "origin": None,
        "ip": None,
        "http": None,
        "route": None,
        "code": None,
        "time": None,
        "user": None,
        "object": None,
        "records": None,
        "text": None,
    }

    # Line should be completely skipped if doesnt start with a date
    date = check_date(line.split(",")[0])
    if not date:
        return (2, None)

    skeletor["date"] = date

    spl = line.split(" ")

    # Determine line type
    lineType = from_line(spl, "type")

    if lineType not in set(get_mapping().keys()):
        return (1, lineType)

    # Based on type, assign all the values to the skeleton

    # Based on type, populate the skeleton
    for k in skeletor:
        # Already assigned date
        if k != "date":
            skeletor[k] = from_line(spl, k, lineType)

    # Add in the type and raw text
    skeletor["type"] = get_mapping()[lineType]["alias"]
    skeletor["text"] = " ".join(spl[get_mapping()["default"]["type"]["index"] + 1 :])

    # If origin is ? in the log, make it null
    if skeletor["origin"] == "?":
        skeletor["origin"] = None

    return (0, skeletor)


def convert_to_sql(full_file_path: str):
    start = time.perf_counter()

    # Set up sql conn
    model = get_table_model()
    BASE.metadata.create_all(ENGINE)
    Session = sessionmaker(bind=ENGINE)
    session = Session()

    # Lines to import
    logLines = []
    skipped = []

    # Calc some stats to show user
    stats = {}

    # Read the selected file

    hide_cursor()
    with gzip.open(full_file_path, "rt") as f:
        with click_spinner.spinner():
            click.echo(
                f"> Reading log file ({os.path.getsize(full_file_path) / 1000000} MB) ",
                nl=False,
            )
            lc = len(f.readlines())
            stats["raw_length"] = lc

            f.seek(0)
        click.echo("\n   Done!")
        with click.progressbar(
            f, label=f"-> Processing raw log lines ({lc})", length=lc
        ) as bar:
            for line in bar:
                status, processed = process_line(line)

                # If processing exits with 1 status, that means the type of the logline was skipped.
                if status == 1:
                    skipped.append(processed)
                    continue
                if status == 2:
                    # 2 is total skip
                    stats["junk_skipped"] = stats.get("junk_skipped", 0) + 1
                    continue

                # Add to the loglines if there were no errors
                logLines.append(processed)

        stats["log_lines_skipped"] = len(skipped)
        # Print out the lines that were skipped based on TYPE
        skipped = condense(skipped)

    logs = []
    with click.progressbar(
        logLines,
        label=f"-> Creating log line objects ({len(logLines)})",
        length=len(logLines),
    ) as ll:
        for l in ll:
            logs.append(model(**l))

    stats["table_len"] = len(logs)

    hide_cursor()
    click.echo("-> Writing to DB ", nl=False)
    with click_spinner.spinner():
        session.add_all(logs)
        session.commit()
    click.echo("\n   Done!")

    show_cursor()
    end = time.perf_counter()
    stats["time"] = end - start
    return (skipped, stats)
